package dev.pfaff.altargui.shader;

import dev.pfaff.glass.opengl.enums.GLVertexAttribPointerType;
import dev.pfaff.glass.opengl.enums.IGLVertexAttribPointerType;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.MemoryLayout;
import java.util.Objects;

public record VertexAttribSpec(String name, IGLVertexAttribPointerType type, boolean normalized, int width, @Nullable MemoryLayout customLayout) {
	public static final VertexAttribSpec PADDING_BYTE = new VertexAttribSpec("padding", GLVertexAttribPointerType.BYTE, false, 1);

	public VertexAttribSpec {
		if (normalized && !(type instanceof GLVertexAttribPointerType)) {
			throw new UnsupportedOperationException("Only GLVertexAttribPointerType can be normalized");
		}
		if (customLayout != null) {
			checkLayoutCompatibility(type, type.layout(), customLayout);
		}
	}

	public VertexAttribSpec(String name, IGLVertexAttribPointerType type, boolean normalized, int width) {
		this(name, type, normalized, width, null);
	}

	public VertexAttribSpec(String name, IGLVertexAttribPointerType type, int width) {
		this(name, type, false, width);
	}

	public VertexAttribSpec withNormalized() {
		return new VertexAttribSpec(name, type, true, width, customLayout);
	}

	public VertexAttribSpec withCustomLayout(MemoryLayout layout) {
		Objects.requireNonNull(layout);
		if (customLayout != null) {
			checkLayoutCompatibility(type, customLayout, layout);
		}
		return new VertexAttribSpec(name, type, normalized, width, layout);
	}

	private static void checkLayoutCompatibility(IGLVertexAttribPointerType type,
												 MemoryLayout base,
												 MemoryLayout custom) {
		if (custom.byteSize() != base.byteSize() || custom.byteAlignment() < base.byteAlignment()) {
			throw new IllegalArgumentException(STR."""
			Incompatible custom layout for \{type} attribute:
				Base layout: \{base}
				Custom layout: \{custom}
			""");
		}
	}
}
